# -*- coding: utf-8 -*-
"""This module supplies some smart utilities"""

import logging
import pyo
import math


def mapToRange(inVal, rangeIn=[0, 1], rangeOut=None, scale="lin", resolution="float"):
    """
    Map a value (inVal) from a range to another range

    Params:

    inVal : numberfloat
    input value

    rangeIn : list
    min and max values expected in input

    rangeOut : list
    min and max value expected in output
    if rangeOut[0] > rangeOut[1] the scale will be reversed accordingly

    scale : str
    scaling mode between rangeIn and rangeOut
    'lin'  - linear
    'exp'  - exponential
    'sqr'  - x**2
    'cub'  - x**3
    'log'  - logarithmic
    'sqrt' - square root (better than 'log')

    resolution : str
    output number resolution
    'float' - default
    'int' - value will be rounded and parsed as an integer
    """

    rangeInMin = min(rangeIn[0],
                     rangeIn[1])
    rangeInMax = max(rangeIn[0],
                     rangeIn[1])
    outVal = float(max(rangeInMin, min(inVal, rangeInMax)))

    if rangeOut is not None:
        rangeOutMin = min(rangeOut[0],
                          rangeOut[1])
        rangeOutMax = max(rangeOut[0],
                          rangeOut[1])
        dIn = rangeInMax - rangeInMin
        dOut = rangeOutMax - rangeOutMin
        normalizedInput = (outVal - rangeInMin) / dIn
        scaleOut = dOut + rangeOutMin
        if scale == 'log':
            outVal = math.log10(normalizedInput * 9 + 1) * scaleOut
        elif scale == 'exp':
            outVal = (math.pow(10, normalizedInput) /
                      9.0 - 1.0 / 9.0) * scaleOut
        elif scale == 'sqr':
            outVal = normalizedInput**2 * scaleOut
        elif scale == 'cub':
            outVal = normalizedInput**3 * scaleOut
        elif scale == 'sqrt':
            outVal = math.sqrt(normalizedInput) * scaleOut
        elif scale == 'lin' or scale is None:
            outVal = normalizedInput * scaleOut

        outVal = max(rangeOutMin, min(outVal, rangeOutMax))

    if resolution == 'int':
        outVal = int(round(outVal))
    return outVal


def mapArgs(method):
    """
    Decorator function : filter a method's argument by applying mapToRange
    Retrieve the arguments from self.exposed_to_osc, which must be a dictionnary
    of the following form :
    self.exposed_to_osc = {
        "methodToExposeAndDecorate": {"rangeIn": [0,1], "rangeOut": [0,127], "scale": "lin", "resolution": "float"},
        # etc
    }


    Usage:

    @mapArgs
    self.methodToExposeAndDecorate(self, value):
        ...
    """

    def setter(self, IN):
        arguments = self.exposed_to_osc[method.__name__]
        OUT = mapToRange(IN, **arguments)
        logging.info("From '%s' mapping '%s %s' to '%s %s' using '%s' scale method" % (method.__name__, IN, arguments.get(
            "rangeIn", [0, 1]), OUT, arguments.get("rangeOut", ""), arguments.get("scale", "lin")))
        return method(self, OUT)

    return setter



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Chainsaw
#
# Copyright (C) 2021 Grégory David <dev@groolot.net>
#
# Chainsaw is free software; you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as
# published by the Free Software Foundation, either version 3 of the
# License, or (at your option) any later version.
#
# Chainsaw is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program. If not, see
# <http://www.gnu.org/licenses/>.

"""Allow user to run Chainsaw as a module."""

# Execute with:
# $ python3 -m chainsaw

import chainsaw

if __name__ == '__main__':
    chainsaw.main()



#!/usr/bin/env python3
import argparse
import logging
import sys
from mididings import *
from mididings.extra.osc import SendOSC

import importlib

class CLI(object):
    """Command Line Interface helper class"""

    def __init__(self):
        """CLI constructor"""
        self.arguments = argparse.Namespace()
        self.generalParser = argparse.ArgumentParser()
        self.oscParser = self.generalParser.add_argument_group('OSC', 'Open Sound Control parameters')
        self.setupParsingCLI()


    def performParsingCLI(self):
        """Do what its name sound"""

        self.arguments = self.generalParser.parse_args()
        screenLoggerFormat = logging.Formatter('%(levelname)s:%(filename)s: %(message)s')
        screenLoggerHandler = logging.StreamHandler(stream=sys.stderr)
        screenLoggerHandler.setFormatter(screenLoggerFormat)
        logging.getLogger().addHandler(screenLoggerHandler)
        logging.getLogger().setLevel(getattr(logging, self.arguments.logLevel))
        logging.debug("Command line arguments: " + repr(self.arguments))
        return self.arguments


    def setupParsingCLI(self):
        """Configure the argument parser with some options"""
        # General options
        self.generalParser.add_argument("-l", "--log-level",
                                        choices=['DEBUG','INFO','WARNING','ERROR','CRITICAL'],
                                        dest="logLevel",
                                        default='WARNING',
                                        help="Select logging output level (default WARNING)")
        self.generalParser.add_argument("-v", "--verbose",
                                        action="store_const",
                                        dest="logLevel",
                                        const="DEBUG",
                                        help="Set the log level to DEBUG and be really verbose")

        # OSC
        self.oscParser.add_argument("-p", "--first-osc-port",
                                    action="store",
                                    type=int,
                                    dest="osc_port",
                                    default=8765,
                                    metavar="FIRST_OSC_PORT_NUMBER",
                                    help="Define the first OSC port to listen to (default 8765) for the first channel. +1 on each additionnal channel.")

        self.oscParser.add_argument("-t", "--osc-target-ip",
                                    action="store",
                                    type=str,
                                    dest="osc_target_ip",
                                    default="127.0.0.1",
                                    metavar="OSC_TARGET_IP",
                                    help="Define the OSC target IPv4 (default to '127.0.0.1')")

        # MIDI to OSC mapping template file
        self.oscParser.add_argument("-f", "--midi2osc-mapping-template-file",
                                    action="store",
                                    type=str,
                                    dest="midi2oscMappingTemplateFilename",
                                    default="midi_chainsaw.mapping_template",
                                    metavar="MAPPING_TEMPLATE_FILE",
                                    help="Define MIDI to OSC mapping template filename, with or without extension .py (default to 'midi_chainsaw.mapping_template')")

        # Number of channels (how many Chainsaw instances)
        self.oscParser.add_argument("-n", "--nchan",
                                    action="store",
                                    type=int,
                                    dest="nchan",
                                    default=1,
                                    metavar="CHANNEL_NUMBER",
                                    help="Define the number of channel to instance")


def main():
    route = {}
    patch = [Print()]
    cli = CLI().performParsingCLI()
    channel_module = importlib.import_module(cli.midi2oscMappingTemplateFilename)
    config(
        backend='jack',
        client_name='midi_chainsaw'
    )
    
    for channel in range(1, cli.nchan + 1):
        route[channel] = channel_module.route
        for cc in route[channel]:
            for address in route[channel][cc]:
                for method in route[channel][cc][address]:
                    logging.info("Register " + address + " " + method)
                    filtre = ChannelFilter(channel) >> CtrlFilter(cc)
                    oscArgs = route[channel][cc][address][method][0:]
                    args = [method]
                    if len(oscArgs) > 0 :
                        """
                        '!' will be replaced by the cc's value normalized between 0 and 1
                        '?' will be replaced by the cc's value (not normalized)
                        If argument list is empty, method is considered as a TRIGGER
                        """
                        for arg in oscArgs:
                            if arg == '!':
                                args.append(lambda ev: ev.value / 127.0)
                            elif arg == '?':
                                args.append(lambda ev: ev.value)
                            else:
                                args.append(arg)
                    else:
                        filtre = filtre >> CtrlValueFilter(lower=1)
                    patch.append(filtre >> SendOSC((cli.osc_target_ip, cli.osc_port + channel - 1), address, *args))

                    logging.debug("Chan" + str(channel) + "/CC#" + str(cc) + " to OSC: " + " ".join([str(address), str(args)]))
    amp_control = ChannelFilter(10) >> (
        CtrlRange(1, 0, 69) >>
        CtrlRange(2, 0, 69) >>
        CtrlRange(3, 0, 69) >>
        CtrlRange(4, 0, 69) >>
        CtrlRange(5, 0, 69) >>
        CtrlRange(6, 0, 69) >>
        CtrlRange(7, 0, 69) >>
        CtrlRange(8, 0, 69)
        )
    patch.append(amp_control)
    run(patch)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""This module supply everything about tracks."""

import logging
import pyo
from chainsaw.osc import OscNode
from chainsaw.effects import EffectChain
from chainsaw.configuration import audioServer, envelopes, monitorManager
from chainsaw.utils import mapArgs


class Track(OscNode):
    """A container representing a TableProcessor and an EffectChain

    WARNING: it does not have to be instanced, just inherited"""

    def __init__(self, name, parent=None):
        """Track constructor

        Params:
        name : str
        Track name

        parent : OscNode
        Parent. Default to None
        """
        OscNode.__init__(self, name, parent=parent)

        self.addChild(EffectChain(name="chain", parent=self, input=self))
        self.children["chain"].addEffect("disto", "Disto")
        self.children["chain"].addEffect("freeverb", "Freeverb")
        self.children["chain"].addEffect("freqshift", "FreqShift")
        self.children["chain"].addEffect("waveguide", "WaveGuide")
        self.children["chain"].setOrder(
            "disto", "freeverb", "freqshift", "waveguide")

        self.chainWetLevel = pyo.Sig(0)
        self.centerController = pyo.Sig(0)
        self.controller = pyo.LFO(freq=0,
                                  mul=1 - self.centerController,
                                  add=self.centerController
                                  )

        self.exposed_to_osc.update(
            {
                "togglePlay": {},
                "setPlay": {},
                "setMul": {"rangeOut": [0, 1], "scale": "sqr"},
                "setWet": {"scale": "sqr"},
                "setMulController": {},
                "setCenterController": {},
                "setFreqController": {"rangeOut": [0, 100], "scale": "cub"},
                "setShapeController": {"rangeIn": [0, 7], "resolution": "int"},
                "setSharpController": {},
                "setMonitor": {}
            }
        )

        self.out_fader = self * (1 - self.chainWetLevel) + \
            self.children["chain"] * self.chainWetLevel
        self.monitor_out = self * \
            (1 - self.chainWetLevel) + \
            self.children["chain"] * self.chainWetLevel
        monitorManager.addInput(input=self.monitor_out)
        self.setPlay(False)
        self.setMul(self.exposed_to_osc["setMul"]["rangeOut"][0])
        self.setWet(0)
        self.setMulController(0)
        self.setCenterController(1)
        self.setFreqController(
            self.exposed_to_osc["setFreqController"]["rangeOut"][0])
        self.setShapeController(
            self.exposed_to_osc["setShapeController"]["rangeIn"][0])
        self.setSharpController(0)
        self.setMonitor(0)

    @mapArgs
    def setMonitor(self, mul):
        """(De)Activate monitoring for this Track

        In facts, it sets quantity sent to the Monitor mixer output.

        Params:
        mul : float
        0 is for deactivate
        1 is for activate
        """
        prefix = "Un-" if int(mul) == 0 else ""
        logging.debug(prefix + "Monitor track '" + self.nodeName + "'")
        self.monitor_out.setMul(mul)

    def getOut(self):
        """Return the output object that contain signal"""
        return self.out_fader

    def out(self, chnl=0, inc=1, dur=0, delay=0):
        """Start processing and send samples to audio output beginning at chnl.

        This method returns self, allowing it to be applied at the object creation.

        Params:
        chnl : int
        Physical output assigned to the first audio stream of the object. Defaults to 0.

        inc : int
        Output channel increment value. Defaults to 1.

        dur : float
        Duration, in seconds, of the object’s activation. The default is 0 and means infinite duration.

        delay : float
        Delay, in seconds, before the object’s activation. Defaults to 0.
        """
        logging.debug("Plays and outputs %s to out:%s with duration:%s and delay:%s" % (
            self.nodeName, chnl, dur, delay))
        self.out_fader.out(chnl, inc, dur, delay)

    def togglePlay(self):
        """
        Toggle Track's state between [playing] and [stopped]
        based on the current playing state
        """
        if self.isPlaying():
            self.stop()
        else:
            self.play()

    def setPlay(self, play=False):
        """
        Set Track's state between [playing] and [stopped]
        """
        if play:
            self.play()
        else:
            self.stop()

    def setController(self, newController):
        """Replace existing controller with a new given one

        This method HAS TO BE OVERRIDEN to comply with subclass controller usage implementation
        """
        if newController is not None:
            logging.debug("Set '%s' controller as %s" %
                          (self.nodeName, newController))
            self.controller = newController
            self.controller.setMul(1 - self.centerController)
            self.controller.setAdd(self.centerController)

    @mapArgs
    def setMul(self, mul):
        """Set the Track multiplication factor through the method mapping (see self.exposed_to_osc)"""
        logging.debug("Set '%s' multiplication factor 'mul' to %s" %
                      (self.nodeName, mul))
        self.out_fader.setMul(mul)

    @mapArgs
    def setWet(self, chainWetLevel):
        """
        Set wet level (0 = dry = input signal only / 1 = wet = processed signal only)

        Params:
        chainWetLevel : float (between 0 and 1)
        """
        logging.debug("Set '%s' chain wet level 'chainWetLevel' to %s" %
                      (self.nodeName, chainWetLevel))
        self.chainWetLevel.setValue(chainWetLevel)

    @mapArgs
    def setMulController(self, mul):
        """Set the Controller multiplication factor through the method mapping (see self.exposed_to_osc)"""
        if hasattr(self.controller, "setMul"):
            logging.debug(
                "Set '%s' controller multiplication factor 'mul' to %s" % (self.nodeName, mul))
            self.controller.setMul(mul)
        else:
            logging.warning("No 'setMul' for %s controller" %
                            (self.controller))

    @mapArgs
    def setCenterController(self, centerValue):
        """Set the Controller center through the method mapping (see self.exposed_to_osc)"""
        logging.debug("Set '%s' controller center value to %s" %
                      (self.nodeName, centerValue))
        self.centerController.setValue(centerValue)

    @mapArgs
    def setFreqController(self, freq):
        """Set the Controller frequency through the method mapping (see self.exposed_to_osc)"""
        if hasattr(self.controller, "setFreq"):
            logging.debug("Set '%s' controller frequency 'freq' to %s" %
                          (self.nodeName, freq))
            self.controller.setFreq(freq)
        else:
            logging.warning("No 'setFreq' for %s controller" %
                            (self.controller))

    @mapArgs
    def setShapeController(self, shape):
        """Set the Controller waveform type through the method mapping (see self.exposed_to_osc)"""
        if hasattr(self.controller, "setType"):
            logging.debug("Set '%s' controller waveform shape 'type' to %s" % (
                self.nodeName, shape))
            self.controller.setType(shape)
        else:
            logging.warning("No 'setType' for %s controller" %
                            (self.controller))

    @mapArgs
    def setSharpController(self, sharp):
        """Set the Controller sharpness factor through the method mapping (see self.exposed_to_osc)"""
        if hasattr(self.controller, "setSharp"):
            logging.debug("Set '%s' controller sharpness factor 'sharp' to %s" % (
                self.nodeName, sharp))
            self.controller.setSharp(sharp)
        else:
            logging.warning("No 'setSharp' for %s controller" %
                            (self.controller))


class LooperTrack(Track, pyo.Looper):
    """A specialized Track and pyo.Looper"""

    def __init__(self, name, table, parent=None, controller=None):
        """LooperTrack constructor

        Params:
        name : str
        Object name

        table : pyo.NewTable
        Input table to be read

        parent : OscNode
        LooperTrack parent that supply OSC commands
        """
        pyo.Looper.__init__(self,
                            table=table,
                            pitch=1,
                            start=0,
                            dur=1.0,
                            xfade=1,  # 1% of the table time
                            mode=1,  # forward
                            xfadeshape=1,  # equal power
                            startfromloop=False,
                            interp=2,
                            autosmooth=False,
                            mul=1,
                            add=0)
        Track.__init__(self, name, parent=parent)

        self.setController(
            self.controller if controller is None else controller)

        self.exposed_to_osc.update(
            {
                "setStart": {"rangeOut": [0, self.table.length], "scale": "cub"},
                "setDur": {"rangeOut": [0, self.table.length], "scale": "cub"},
                "setMode": {"rangeIn": [1, 3], "resolution": "int"}
            }
        )

        self.setStart(0)
        self.setDur(self.table.length)
        self.setMode(1)

    def setController(self, newController):
        """Replace existing controller with a new given one

        OVERRIDEN to comply with subclass controller usage implementation
        """
        Track.setController(self, newController)
        logging.debug("Set '%s' pitch controller to %s" %
                      (self.nodeName, newController))
        self.setPitch(self.controller)

    @mapArgs
    def setStart(self, position):
        """Mapped wrapper to pyo.Looper.setStart()

        Params:
        position : float
        Starting point, in seconds, of the loop, updated only once per loop cycle.
        """
        logging.debug("Set '%s' LooperTrack start position 'start' to %s seconds" % (
            self.nodeName, position))
        pyo.Looper.setStart(self, position)

    @mapArgs
    def setDur(self, duration):
        """Mapped wrapper to pyo.Looper.setDur()

        Params:
        duration : float
        Duration, in seconds, of the loop, updated only once per loop cycle.
        """
        logging.debug("Set '%s' LooperTrack duration 'dur' to %s seconds" % (
            self.nodeName, duration))
        pyo.Looper.setDur(self, duration)

    @mapArgs
    def setMode(self, mode):
        """
        Loop modes.

        0 = no loop
        1 = forward
        2 = backward
        3 = back-and-forth
        """
        modename = ["no loop", "forward", "backward", "back-and-forth"]
        logging.debug("Set '%s' LooperTrack 'mode' to %s (%s)" %
                      (self.nodeName, mode, modename[mode]))
        pyo.Looper.setMode(self, mode)


class ParticleTrack(Track, pyo.Particle):
    """A specialized Track and pyo.Particle"""

    def __init__(self, name, table, parent=None, controller=None):
        """ParticleTrack constructor

        Params:
        name : str
        Object name

        table : pyo.NewTable
        Input table to be read

        parent : OscNode
        ParticleTrack parent that supply OSC commands
        """
        pyo.Particle.__init__(self,
                              table=table,
                              env=envelopes[5],
                              dens=1,
                              pitch=1,
                              pos=0,
                              dur=0,
                              dev=0.1,
                              pan=0.5,
                              chnls=1
                              )
        Track.__init__(self,
                       name,
                       parent=parent)

        self.setController(
            self.controller if controller is None else controller)

        self.exposed_to_osc.update(
            {
                "setEnv": {"rangeIn": [0, len(envelopes)], "resolution": "int"},
                "setDens": {"rangeOut": [1 / self.table.length, 50], "scale": "cub"},
                "setPos": {"rangeOut": [0, self.table.size], "scale": "cub"},
                "setDur": {"rangeOut": [0, self.table.length-0.00001], "scale": "cub"},
                "setDev": {}
            }
        )

        self.setEnv(0)
        self.setDens(self.exposed_to_osc["setDens"]["rangeOut"][0])
        self.setPos(0.34)
        self.setDur(self.table.length / 16)
        self.setDev(0.24)

    def setController(self, newController):
        """Replace existing controller with a new given one

        OVERRIDEN to comply with subclass controller usage implementation
        """
        Track.setController(self, newController)
        logging.debug("Set '%s' 'pitch' to %s" %
                      (self.nodeName, newController))
        self.setPitch(self.controller)

    @mapArgs
    def setEnv(self, env):
        """Mapped wrapper to pyo.Particle.setEnv()

        Params:
        env : int
        configuration.envelopes[] Index reference to a Table containing the grain envelope

        WARNING: some envelopes make glitches due to their shapes
        See: pyo_0.7.9-doc/api/classes/tables.html
        """
        logging.debug("Set '%s' ParticleTrack envelope 'env' to %s (%s)" % (
            self.nodeName, env, envelopes[env]))
        pyo.Particle.setEnv(self, envelopes[env])

    @mapArgs
    def setDens(self, density):
        """Mapped wrapper to pyo.Particle.setDens()

        Params:
        density : float
        Density of grains per second.
        """
        logging.debug("Set '%s' ParticleTrack density 'dens' to %s" %
                      (self.nodeName, density))
        pyo.Particle.setDens(self, density)

    @mapArgs
    def setPos(self, position):
        """Mapped wrapper to pyo.Particle.setPos()

        Params:
        position : int
        Pointer position, in samples, in the waveform table.
        Each grain sampled the current value of this stream at the beginning
        of its envelope and hold it until the end of the grain.
        """
        logging.debug("Set '%s' ParticleTrack position 'pos' to %s" %
                      (self.nodeName, position))
        pyo.Particle.setPos(self, position)

    @mapArgs
    def setDur(self, duration):
        """Mapped wrapper to pyo.Particle.setDur()

        Params:
        duration : float
        Duration, in seconds, of the grain. Each grain sampled the current value
        of this stream at the beginning of its envelope and hold it until the
        end of the grain.
        """
        logging.debug("Set '%s' ParticleTrack duration 'dur' to %s" %
                      (self.nodeName, duration))
        pyo.Particle.setDur(self, duration)

    @mapArgs
    def setDev(self, deviation):
        """Mapped wrapper to pyo.Particle.setDev()

        Params:
        deviation : float
        Maximum deviation of the starting time of the grain, between 0 and 1 (relative
        to the current duration of the grain).
        Each grain sampled the current value of this stream at the beginning of
        its envelope and hold it until the end of the grain.
        """
        logging.debug("Set '%s' ParticleTrack maximum deviation 'dev' to %s" % (
            self.nodeName, deviation))
        pyo.Particle.setDev(self, deviation)


if __name__ == "__main__":
    from chainsaw.osc import OscRootNode
    from chainsaw.configuration import commandLine
    oscRouter = OscRootNode("tracks", oscPort=commandLine.port)

    table = pyo.SndTable(pyo.SNDS_PATH + "/transparent.aif")
    table.length = table.getDur()  # mimic NewTable length attribute

    particle = ParticleTrack('particle', t)
    oscRouter.addChild(particle)
    particle.play()
    particle.centerController.ctrl(title="particle pitch center")
    # particle.controller.ctrl(title="pitch controller")
    mixer = pyo.Mix(particle, 2).out()

    looper = LooperTrack('looper', t)
    oscRouter.addChild(looper)
    looper.setMulController(.5)
    looper.setFreqController(.5)
    looper.setController(pyo.BrownNoise())
    # Next do not do anything, because there is no Freq attribute in pyo.BrownNoise, but print information
    looper.setFreqController(.5)
    # looper.out()
    # looper.centerController.ctrl()
    # looper.controller.ctrl()
    # looper.ctrl()

    audioServer.start()
    audioServer.gui(locals())



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""This module supply Effect, EffectChain and specialized effects.

Speciqlized are: Disto, Freeverb and FreqShift (list to be scaled up).

"""

import logging
import pyo
from chainsaw.osc import OscNode
from chainsaw.configuration import audioInputs
from chainsaw.utils import mapArgs


class Effect(OscNode):
    """Pyo effect wrapper with dry/wet setting support"""

    def __init__(self, name, parent, input=audioInputs[0]):
        """
        Effect constructor
        Make sure "effectType" matches a valid Pyo effect class
        Instanciate a FaderInput to act as the effect's input wrapper
        Instanciate the effect processor which takes the FaderInput as input
        Inherit from pyo.Mixer to output the audio stream
        and handle Dry/Wet setting between these two

        Params:
        name : str
        effectType : str
        input : PyoObject
        """
        OscNode.__init__(self, name, parent)

        self.in_fader = pyo.InputFader(input)
        self.wetLevel = pyo.Sig(0)
        self.out_fader = self.in_fader * \
            (1-self.wetLevel) + self * self.wetLevel
        logging.debug("Creates a new effect named '%s' as child of '%s'" % (
            name, getattr(parent, "nodeName", "ROOT")))
        self.exposed_to_osc.update(
            {
                "setWet": {"scale": "sqr"}
            }
        )

    def getOut(self):
        """Return the correct output based on the inner mixing mechanism"""
        return self.out_fader

    def out(self, chnl=0, inc=1, dur=0, delay=0):
        """Start processing and send samples to audio output beginning at chnl.

        This method returns self, allowing it to be applied at the object creation.

        Params:
        chnl : int
        Physical output assigned to the first audio stream of the object. Defaults to 0.

        inc : int
        Output channel increment value. Defaults to 1.

        dur : float
        Duration, in seconds, of the object’s activation. The default is 0 and means infinite duration.

        delay : float
        Delay, in seconds, before the object’s activation. Defaults to 0.
        """
        self.out_fader.out(chnl, inc, dur, delay)

    @mapArgs
    def setWet(self, wetLevel):
        """
        Set wet level (0 = dry = input signal only / 1 = wet = processed signal only)

        Params:
        wetLevel : float (between 0 and 1)
        """
        logging.info("Set '%s' effect wet level 'wetLevel' to %s" %
                     (self.nodeName, wetLevel))
        logging.debug("%s.wetLevel.setValue(%s)" % (self.nodeName, wetLevel))
        self.wetLevel.setValue(wetLevel)

    def setInput(self, newInput):
        """
        Set Effect's input

        Params:
        newInput : PyoObject
        """
        logging.info("Set '%s' audio input signal 'input' to %s" %
                     (self.nodeName, newInput))
        logging.debug("%s.in_fader.setInput(%s)" % (self.nodeName, newInput))
        self.in_fader.setInput(newInput)


class Disto(Effect, pyo.Disto):
    """Specialized effect as Disto"""

    def __init__(self, name, parent, input=audioInputs[0]):
        """Disto constructor"""
        pyo.Disto.__init__(self, input=audioInputs[0])
        Effect.__init__(self, name, parent, input)
        pyo.Disto.setInput(self, self.in_fader)
        self.exposed_to_osc.update(
            {
                "setDrive": {}
            }
        )

    @mapArgs
    def setDrive(self, drive):
        """
        Set disto drive amount

        Params:
        drive : float [0..1]
           Drive amount normalized
        """
        logging.info("Set '%s' Disto drive amount 'drive' to %s" %
                     (self.nodeName, drive))
        logging.debug("%s.setDrive(%s)" % (self.nodeName, drive))
        pyo.Disto.setDrive(self, drive)


class Freeverb(Effect, pyo.Freeverb):
    """Specialized effect as Freeverb"""

    def __init__(self, name, parent, input=audioInputs[0]):
        """Freeverb constructor"""
        pyo.Freeverb.__init__(
            self, input=audioInputs[0], size=0, damp=0.5, bal=1.0)
        Effect.__init__(self, name, parent, input)
        pyo.Freeverb.setInput(self, self.in_fader)
        self.exposed_to_osc.update(
            {
                "setSize": {}
            }
        )

    @mapArgs
    def setSize(self, size):
        """
        Set reverb room size

        Params:
        size : float [0..1]
           Reverb room size
        """
        logging.info("Set '%s' Freeverb room size 'size' to %s" %
                     (self.nodeName, size))
        logging.debug("%s.setSize(%s)" % (self.nodeName, size))
        pyo.Freeverb.setSize(self, size)


class FreqShift(Effect, pyo.FreqShift):
    """Specialized effect as FreqShift"""

    def __init__(self, name, parent, input=audioInputs[0]):
        """FreqShift constructor"""
        pyo.FreqShift.__init__(self, input=audioInputs[0])
        Effect.__init__(self, name, parent, input)
        pyo.FreqShift.setInput(self, self.in_fader)
        self.exposed_to_osc.update(
            {
                "setShift": {"rangeOut": [0, 1000]}
            }
        )

    @mapArgs
    def setShift(self, freq):
        """
        Set FreqShift frequency shift

        Params:
        freq : float [0..1000]
           Frequency shifting value
        """
        logging.info("Set '%s' FreqShift frequency shifting value 'shift' to %s" % (
            self.nodeName, freq))
        logging.debug("%s.setShift(%s)" % (self.nodeName, freq))
        pyo.FreqShift.setShift(self, freq)


class WaveGuide(Effect, pyo.AllpassWG):
    """Specialized effect as AllpassWG"""

    def __init__(self, name, parent, input=audioInputs[0]):
        """AllpassWG constructor"""
        self.freqFactor = pyo.Randi(
            min=.95, max=1.05, freq=[.145, .2002, .1055, .071])
        pyo.AllpassWG.__init__(self, input=input, freq=self.freqFactor*[
                               74.99, 75, 75.01, 75.1], feed=1, detune=pyo.Randi(min=.5, max=1.0, freq=[.022, .13, .155, .171]))
        Effect.__init__(self, name, parent, input)
        pyo.AllpassWG.setInput(self, self.in_fader)

        self.exposed_to_osc.update(
            {
                "setFreq": {"rangeOut": [0, 10000], "scale": "cub"}
            }
        )

    @mapArgs
    def setFreq(self, factor):
        """Set frequency factor to drive harmonics

        Params:
        factor : float [0..10000]
           Frequency factor value
        """
        logging.info("Set '%s' AllpassWG frequency factor value to '%s'" %
                     (self.nodeName, factor))
        logging.debug("%s.setFreq(%s)" % (self.nodeName, factor))
        self.freqFactor = pyo.Randi(min=.95, max=1.05, freq=[
                                    factor*.145, factor*.2002, factor*.1055, factor*.071])
        pyo.AllpassWG.setFreq(self, self.freqFactor*[74.99, 75, 75.01, 75.1])


class EffectChain(OscNode, pyo.PyoObject):
    """Contains multiple effects and manage their order"""

    def __init__(self, name, parent, children=None, order=None, input=audioInputs[0]):
        """
        EffectChain constructor
        """
        OscNode.__init__(self, name, parent, children)
        pyo.PyoObject.__init__(self)

        self.effectClasses = {
            "Disto": Disto,
            "Freeverb": Freeverb,
            "FreqShift": FreqShift,
            "WaveGuide": WaveGuide
        }

        self.order = order if order is not None else []
        self.in_fader = pyo.InputFader(input)

        self.out_fader = pyo.InputFader(audioInputs[0])
        self._base_objs = self.out_fader.getBaseObjects()

    def setInput(self, newInput):
        """
        Set Chain's input

        Params:
        newInput : PyoObject
        """
        logging.debug("Set '%s' audio input signal 'input' to %s" %
                      (self.nodeName, newInput))
        self.in_fader.setInput(newInput)

    def addEffect(self, name, effectType):
        """
        Add an Effect object to the chain

        Params:
        name : str
        effectType : str
        """
        logging.debug("Parent '%s' wants a new child effect '%s' named '%s'" % (
            self.nodeName, effectType, name))
        if effectType in self.effectClasses.keys():
            self.addChild(self.effectClasses[effectType](name, self))
            self.order.append(name)
            self.refreshPatch()

    def removeEffect(self, name):
        """
        Remove an Effect object from the chain

        Params:
        name : name
        """
        if name in self.children.keys():
            logging.info("Remove from '%s' effect '%s'" %
                         (self.nodeName, name))
            del self.children[name]
            del self.order[self.order.index(name)]
            self.refreshPatch()

    def setOrder(self, *newOrder):
        """
        Set a new patch order for the effects in the chain

        Params:
        *newOrder : multiple args (str)
        """
        logging.info("Set '%s' chain order to '%s'" %
                     (self.nodeName, newOrder))
        self.order = newOrder
        self.refreshPatch()

    def refreshPatch(self):
        """
        Set the effect's inputs according to the chain's order
        Special cases:
        First effect's input = chainInput
        Chain's Mix (self) must replace his input with the chain's last effect
        """
        logging.debug("Refresh '%s' effect chain patch" % (self.nodeName))
        previousStream = self.in_fader
        cursor = 0
        for item in self.order:
            if item in self.children.keys():
                if cursor == 0:
                    self.children[item].setInput(previousStream)
                else:
                    self.children[item].setInput(previousStream.getOut())

                cursor += 1
                previousStream = self.children[item]

        if previousStream != self.in_fader:
            self.out_fader.setInput(previousStream.getOut())

    def play(self, dur=0, delay=0):
        """Start processing without sending samples to output. This method is called automatically at the object creation.

        This method returns self, allowing it to be applied at the object creation.

        Params:
        dur : float
        Duration, in seconds, of the object’s activation. The default is 0 and means infinite duration.

        delay : float
        Delay, in seconds, before the object’s activation. Defaults to 0.
        """
        self.out_fader.play(dur, delay)
        return pyo.PyoObject.play(self, dur, delay)

    def stop(self):
        """Stop processing.

        This method returns self, allowing it to be applied at the object creation.
        """
        self.out_fader.stop()
        return pyo.PyoObject.stop(self)

    def out(self, chnl=0, inc=1, dur=0, delay=0):
        """Start processing and send samples to audio output beginning at chnl.

        This method returns self, allowing it to be applied at the object creation.

        Params:
        chnl : int
        Physical output assigned to the first audio stream of the object. Defaults to 0.

        inc : int
        Output channel increment value. Defaults to 1.

        dur : float
        Duration, in seconds, of the object’s activation. The default is 0 and means infinite duration.

        delay : float
        Delay, in seconds, before the object’s activation. Defaults to 0.
        """
        self.out_fader.play(dur, delay)
        return pyo.PyoObject.out(self, chnl, inc, dur, delay)


if __name__ == "__main__":
    oneEffect = FreqShift("freqshiftTest", None)
    oneEffect.setInput(audioInputs[1])
    oneEffect.setWet(.8)
    oneEffect.ctrl()
    oneEffect.out()

    oneChain = EffectChain("chainTest", None)
    oneChain.setInput(audioInputs[1])
    oneChain.addEffect("fatDisto", "Disto")
    oneChain.children["fatDisto"].ctrl()
    oneChain.addEffect("fatVerb", "Freeverb")
    oneChain.children["fatVerb"].ctrl()
    oneChain.setOrder("fatVerb", "fatDisto")
    oneChain.out()

    from chainsaw.osc import OscRootNode
    from chainsaw.configuration import commandLine
    oscRouter = OscRootNode("effect", oscPort=commandLine.port)
    oscRouter.addChild(oneEffect)
    oscRouter.addChild(oneChain)
    oscSender = pyo.OscDataSend(
        "sf", commandLine.port, "/effect/chainTest/fatDisto")
    oscSender.send(["setWet", 1])
    oscSender = pyo.OscDataSend(
        "sf", commandLine.port, "/effect/chainTest/fatVerb")
    oscSender.send(["setWet", 1])
    oscSender = pyo.OscDataSend(
        "sf", commandLine.port, "/effect/chainTest/fat*")
    oscSender.send(["setWet", .5])
    oscSender = pyo.OscDataSend("sf", commandLine.port, "/effect/*/*")
    oscSender.send(["setWet", .2])
    oscSender = pyo.OscDataSend(
        "sf", commandLine.port, "/effect/fre[Qq]shi?tTest")
    oscSender.send(["setWet", 1])

    from chainsaw.configuration import audioServer
    audioServer.start()
    audioServer.gui(locals())



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""This module encapsulate the OSC mechanism"""

import logging
import pyo
import re


class OscNode(object):
    """Represent an individual node in the OSC tree"""

    def __init__(self, name, parent=None, children=None):
        """
        OscNode constructor
        Set the node's name that will be use as an identifier

        Params:
        name : str
        """
        self.nodeName = str(name)
        self.parent = parent
        self.children = children if children is not None else {}
        self.exposed_to_osc = {}

    def addChild(self, childNode):
        """
        Add a child to the node, store it children against its nodeName

        Params:
        childNode : OscNode
        """
        if not hasattr(childNode, "nodeName"):
            raise TypeError('ERROR: could not add child %s to node %s (not a valid OscNode instance)' % (
                str(childNode), self.nodeName))

        self.removeChild(childNode.nodeName)
        logging.debug("Parent '%s' has a new child: '%s'" %
                      (self.nodeName, childNode.nodeName))
        self.children[childNode.nodeName] = childNode

    def removeChild(self, childName):
        """
        Remove a child from the node

        Params:
        childName : str
        """
        if childName in self.children:
            logging.debug("Parent '%s' lost child: '%s'" %
                          (self.nodeName, childName))
            del self.children[childName]

    def stringMatch(self, regExpPattern, subject):
        """
        Shorthand for regexp pattern matching
        Returns a boolean

        Params:
        regExpPattern : str
        subject : str
        """
        pattern = re.compile("^" + regExpPattern + "$")
        match = pattern.match(subject)
        matched = match != None and len(match.string) > 0
        match_string = "Match" if matched else "Mismatch"
        logging.debug("%s node '%s' comparing '%s' to '%s'" %
                      (match_string, self.nodeName, regExpPattern, subject))
        return matched

    def receiveOsc(self, path, *args):
        """
        Receive OSC message and check if the node should interpret it
        """
        if self.stringMatch(path[0], self.nodeName):
            logging.debug("Node '%s' process OSC path: %s" %
                          (self.nodeName, path))
            self.processOsc(path, *args)

    def processOsc(self, path, *args):
        """
        Process received osc message

        Params:
        path : list (of node names)
        *args : anything?
        """
        if len(path) == 1:
            # The message has reach its target
            # args[0] = method to execute
            # args[1:] = args to pass
            methodName = self.lastMethod = args[0]
            if methodName in self.exposed_to_osc:
                methodArgs = args[1:]
                getattr(self, methodName, self.processFallback)(*methodArgs)
        else:
            # pass the message to the next target
            newPath = path[1:]

            for name in self.children:
                self.children[name].receiveOsc(newPath, *args)

    def processFallback(self, *args):
        """Process OSC fallback

        When a processOsc do not find any method to call, this one is called and
        prints a nice warning message
        """
        logging.warning("'%s' is not defined in '%s' node. Args were: %s" % (
            self.lastMethod, self.nodeName, str(args)))

    def debug(self, *args):
        """Debugging"""
        logging.debug("'%s' node receives '%s' args" %
                      (self.nodeName, str(args)))


class OscRootNode(OscNode):
    """Root node of the OSC tree"""

    def __init__(self, name, children=None, oscPort=None):
        """
        OscRootNode constructor
        Instanciate an OSC server listening on port "oscPort"
        Set a list of children whom OSC messages will be transfered to

        Params:
        name : str
        oscPort : int
        """
        OscNode.__init__(self, name, children=children)
        logging.info("'%s' creates an OSC server listening on port:'%s'" %
                     (self.nodeName, oscPort))
        self.server = pyo.OscListener(self.normalizeOsc, oscPort)
        logging.debug("'%s' starts its OSC server" % (self.nodeName))
        self.server.start()

    def normalizeAddress(self, address):
        """
        Convert OSC 1.1 compliant address to regexp pattern standards
        Escape ^, $ (start/end of string delimiters) and \ (escape char)
        ?           -> .?
        [!a-Z]      -> [^a-Z]
        {foo,bar}   -> (foo|bar)

        Params:
        address : str
        """
        patterns = {
            r"\?": ".",
            r"\*": ".*",
            r"\[!([^\]]*)\]": r"[^\1]",
            r"\$": r"\$",
            r"\^": r"\^",
            r"\\": r"\\"
        }

        for pattern, repl in patterns.items():
            normAddress = re.sub(pattern, repl, address)

        def transliteration(match):
            s = match.group(0)
            s = s.replace("{", "(")
            s = s.replace("}", ")")
            s = s.replace(",", "|")
            return s

        normAddress = re.sub(re.compile(
            r"\{[^\}]*\}"), transliteration, normAddress)

        logging.debug("'%s' normalized OSC address from '%s' to '%s'" %
                      (self.nodeName, address, normAddress))
        return normAddress

    def normalizeOsc(self, address, *args):
        """
        Normalise OSC message and pass it to receiveOsc method
        Split the address into a path array whose items represent the nodes to traverse
        """

        normAddress = self.normalizeAddress(address)

        if normAddress[0] == "/":
            normAddress = normAddress[1:]
        if normAddress[-1] == "/":
            normAddress = normAddress[0:len(normAddress)-2]

        path = normAddress.split("/")
        logging.debug("'%s' computed OSC path from '%s' to '%s'" %
                      (self.nodeName, address, path))
        self.receiveOsc(path, *args)

    def dumpOscTree(self):
        """Prints the OSC tree from this root node"""
        logging.info("==== BEGIN DUMP OSC TREE ====")
        startLevel = 0

        def traverseTree(child, level):
            tabs = level * 2 * " "
            logging.info(tabs + child.nodeName)
            if len(child.exposed_to_osc) > 0:
                tabs += " "
                for childFunction in sorted(child.exposed_to_osc.keys()):
                    logging.info(tabs + childFunction + "()")
            if len(child.children) > 0:
                keys = sorted(child.children)
                for childName in keys:
                    traverseTree(child.children[childName], level + 1)

        traverseTree(self, startLevel)
        logging.info("==== END DUMP OSC TREE ====")


if __name__ == "__main__":
    from chainsaw.configuration import commandLine
    oscRouter = OscRootNode("osc", oscPort=commandLine.port)
    oscRouter.addChild(OscNode("nodeOne"))
    oscRouter.addChild(OscNode("nodeTwo"))
    oscRouter.children["nodeTwo"].addChild(OscNode("nodeTwoSubOne"))
    oscRouter.children["nodeTwo"].addChild(OscNode("nodeTwoSubTwo"))
    oscRouter.addChild(OscNode("nodeThree"))
    oscRouter.addChild(OscNode("nodeFour"))
    oscRouter.children["nodeFour"].addChild(OscNode("nodeFourSubOne"))
    oscSender = pyo.OscDataSend("sfiTFdN", commandLine.port, "/osc/node*")
    oscSender.send(["debug", 1.0, 2, True, False, 1234.5678, None])
    oscRouter.dumpOscTree()

    from configuration import audioServer
    audioServer.start()
    audioServer.gui(locals())



# -*- coding: utf-8 -*-
"""Configuration module to share data globally inside pyoChainsaw"""
import pyo
import chainsaw.cli as cli
from chainsaw.monitor import MonitorManager

"""
Parse CLI arguments
"""
commandLine = cli.CLI().performParsingCLI()

if commandLine.listPortaudioDevices:
    pyo.pa_list_devices()
    exit(0)


"""
Instanciate the audio server
Once the pyo.Server is booted, whatever signal can
be added to the global stream.
"""
jackname = "{!s}-{!s}".format(commandLine.jackname, commandLine.port)
audioServer = pyo.Server(audio=commandLine.audio,
                         sr=commandLine.samplingRate,
                         jackname=jackname,
                         ichnls=1,
                         nchnls=2)  # + 1 for monitoring
pyoVerbosityMapping = {
    'DEBUG': 8,
    'INFO': 4,
    'WARNING': 2,
    'ERROR': 1,
    'CRITICAL': 1,
}
audioServer.setVerbosity(pyoVerbosityMapping[commandLine.logLevel])
audioServer.deactivateMidi()

if commandLine.audio == "portaudio":
    audioServer.setInOutDevice(commandLine.portaudioDevice)
elif commandLine.audio == "jack":
    audioServer.setJackAuto(xin=False, xout=False)

audioServer.boot()


"""
Construct a tuple containing the input audio stream
"""
audioInputs = (pyo.Sig(0), pyo.Input(0))  # Set to tuple to avoid mutation


"""
Base Envelopes
"""
envelopes = [
    pyo.ChebyTable(),
    pyo.CosLogTable(),
    pyo.CosTable(),
    pyo.CurveTable(),
    pyo.ExpTable(),
    pyo.HannTable(),
    pyo.HarmTable(),
    pyo.LinTable(),
    pyo.LogTable(),
    pyo.ParaTable(),
    pyo.PartialTable(),
    pyo.SawTable(),
    pyo.SincTable(),
    pyo.SquareTable(),
    pyo.AtanTable()
]

"""
Monitors Manager
"""
monitorManager = MonitorManager('Monitor', audioServer.getNchnls() - 1)



"""
Route patch dictionnary, must be of this form:
    route = {
        control_change_number: {
            "/osc/address": {
                "method_name": [arg1, args2, "!"],
                "method_name2": ["?"],
                "triggered_method_name3": []
            }
        }
    }

Where:
    "!" will be replaced by the cc's value normalized between 0 and 1
    "?" will be replaced by the cc's value (not normalized)
    If argument list is empty, method is considered as a TRIGGER
"""
route = {
    10: {
        "/chainsaw/instrument/looperTrack": {
            "setMul": ["!"]
        }
    },
    11: {
        "/chainsaw/instrument/looperTrack": {
            "setStart": ["!"]
        }
    },
    12: {
        "/chainsaw/instrument/looperTrack": {
            "setMulController": ["!"]
        }
    },
    13: {
        "/chainsaw/instrument/looperTrack": {
            "setFreqController": ["!"]
        }
    },
    14: {
        "/chainsaw/instrument/looperTrack": {
            "setSharpController": ["!"]
        }
    },
    15: {
        "/chainsaw/instrument": {
            "snapShot": []
        }
    },
    16: {
        "/chainsaw/instrument/looperTrack": {
            "setPlay": ["!"]
        }
    },
    20: {
        "/chainsaw/instrument/Monitor": {
            "setMul": ["!"]
        }
    },
    21: {
        "/chainsaw/instrument/looperTrack": {
            "setDur": ["!"]
        }
    },
    22: {
        "/chainsaw/instrument/looperTrack": {
            "setCenterController": ["!"]
        }
    },
    23: {
        "/chainsaw/instrument/looperTrack": {
            "setShapeController": ["?"]
        }
    },
    24: {
        "/chainsaw/instrument/looperTrack": {
            "setMode": ["?"]
        }
    },
    25: {
        "/chainsaw/instrument": {
            "debug": ["Unnused button, value: ","?"]
        }
    },
    26: {
        "/chainsaw/instrument/looperTrack": {
            "setMonitor": ["!"]
        }
    },
    # Effect Chain for looperTrack
    30: {
        "/chainsaw/instrument/looperTrack": {
            "setWet": ["!"]
        }
    },
    ## Disto
    31: {
        "/chainsaw/instrument/looperTrack/chain/disto": {
            "setDrive": ["!"]
        }
    },
    41: {
        "/chainsaw/instrument/looperTrack/chain/disto": {
            "setWet": ["!"]
        }
    },
    ## Freeverb
    32: {
        "/chainsaw/instrument/looperTrack/chain/freeverb": {
            "setSize": ["!"]
        }
    },
    42: {
        "/chainsaw/instrument/looperTrack/chain/freeverb": {
            "setWet": ["!"]
        }
    },
    ## FreqShift
    33: {
        "/chainsaw/instrument/looperTrack/chain/freqshift": {
            "setShift": ["!"]
        }
    },
    43: {
        "/chainsaw/instrument/looperTrack/chain/freqshift": {
            "setWet": ["!"]
        }
    },
    ## WaveGuide
    34: {
        "/chainsaw/instrument/looperTrack/chain/waveguide": {
            "setFreq": ["!"]
        }
    },
    44: {
        "/chainsaw/instrument/looperTrack/chain/waveguide": {
            "setWet": ["!"]
        }
    },
    50: {
        "/chainsaw/instrument/particleTrack": {
            "setMul": ["!"]
        }
    },
    51: {
        "/chainsaw/instrument/particleTrack": {
            "setPos": ["!"]
        }
    },
    52: {
        "/chainsaw/instrument/particleTrack": {
            "setMulController": ["!"]
        }
    },
    53: {
        "/chainsaw/instrument/particleTrack": {
            "setFreqController": ["!"]
        }
    },
    54: {
        "/chainsaw/instrument/particleTrack": {
            "setDev": ["!"]
        }
    },
    55: {
        "/chainsaw/instrument": {
            "snapShot": []
        }
    },
    56: {
        "/chainsaw/instrument/particleTrack": {
            "setPlay": ["!"]
        }
    },
    60: {
        "/chainsaw/instrument/particleTrack": {
            "setDens": ["!"]
        }
    },
    61: {
        "/chainsaw/instrument/particleTrack": {
            "setDur": ["!"]
        }
    },
    62: {
        "/chainsaw/instrument/particleTrack": {
            "setCenterController": ["!"]
        }
    },
    63: {
        "/chainsaw/instrument/particleTrack": {
            "setShapeController": ["?"]
        }
    },
    64: {
        "/chainsaw/instrument/particleTrack": {
            "setEnv": ["?"]
        }
    },
    65: {
        "/chainsaw/instrument": {
            "debug": ["Unnused button, value: ","?"]
        }
    },
    66: {
        "/chainsaw/instrument/particleTrack": {
            "setMonitor": ["!"]
        }
    },
    # Effect Chain for particleTrack
    70: {
        "/chainsaw/instrument/particleTrack": {
            "setWet": ["!"]
        }
    },
    ## Disto
    71: {
        "/chainsaw/instrument/particleTrack/chain/disto": {
            "setDrive": ["!"]
        }
    },
    81: {
        "/chainsaw/instrument/particleTrack/chain/disto": {
            "setWet": ["!"]
        }
    },
    ## Freeverb
    72: {
        "/chainsaw/instrument/particleTrack/chain/freeverb": {
            "setSize": ["!"]
        }
    },
    82: {
        "/chainsaw/instrument/particleTrack/chain/freeverb": {
            "setWet": ["!"]
        }
    },
    ## FreqShift
    73: {
        "/chainsaw/instrument/particleTrack/chain/freqshift": {
            "setShift": ["!"]
        }
    },
    83: {
        "/chainsaw/instrument/particleTrack/chain/freqshift": {
            "setWet": ["!"]
        }
    },
    ## WaveGuide
    74: {
        "/chainsaw/instrument/particleTrack/chain/waveguide": {
            "setFreq": ["!"]
        }
    },
    84: {
        "/chainsaw/instrument/particleTrack/chain/waveguide": {
            "setWet": ["!"]
        }
    }
}



from chainsaw.osc import OscNode
from chainsaw.utils import mapArgs
import pyo


class MonitorManager(OscNode, pyo.Mixer):
    """
    Monitors Manager
    """

    def __init__(self, name, chnl):
        OscNode.__init__(self, name)
        pyo.Mixer.__init__(self, outs=1, chnls=1)
        pyo.Mixer.out(self, chnl)

        self.exposed_to_osc.update(
            {
                "setMul": {"scale": "sqr"}
            }
        )

    def addInput(self, input):
        key = pyo.Mixer.addInput(self, None, input)
        self.setAmp(key, 0, 1)

    @mapArgs
    def setMul(self, mul):
        pyo.Mixer.setMul(self, mul)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import logging
import pyo
from chainsaw.configuration import jackname, audioInputs, audioServer, commandLine, monitorManager
from chainsaw.osc import OscRootNode

# Global name
__program__ = "chainsaw"
__version__ = '0.3.2'
__author__ = 'Grégory David'
__author_email__ = 'dev@groolot.net'
__license__ = 'GPLv3'


def main():
    """
    Main program entry
    """
    from chainsaw.instruments import Instrument

    oscRoot = OscRootNode(commandLine.jackname, oscPort=commandLine.port)
    instrument = Instrument("instrument")
    instrument.out()

    oscRoot.addChild(instrument)
    oscRoot.addChild(monitorManager)
    oscRoot.dumpOscTree()
    audioServer.start()
    audioServer.gui(
        locals=locals() if logging.getLogger().getEffectiveLevel() == logging.DEBUG else None,
        title=jackname,
        exit=False
    )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# MIDI-Chainsaw
#
# Copyright (C) 2021 Grégory David <dev@groolot.net>
#
# MIDI-Chainsaw is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public License
# as published by the Free Software Foundation, either version 3 of
# the License, or (at your option) any later version.
#
# MDIDI-Chainsaw is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program. If not, see
# <http://www.gnu.org/licenses/>.

"""Allow user to run MIDI-Chainsaw as a module."""

# Execute with:
# $ python -m midi_chainsaw

import midi_chainsaw

if __name__ == '__main__':
    midi_chainsaw.main()



# -*- coding: utf-8 -*-
"""Command Line Interface module"""

import argparse
import logging
import sys


class CLI(object):
    """Command Line Interface helper class"""

    def __init__(self):
        """CLI constructor"""
        self.arguments = argparse.Namespace()
        self.generalParser = argparse.ArgumentParser()
        self.audio_group = self.oscParser.add_mutually_exclusive_group()
        self.portaudioParser = self.audio_group.add_argument_group(
            'PORTAUDIO', 'Define specific options to Portaudio audio driver')
        self.jackParser = self.audio_group.add_argument_group(
            'JACK', 'Jack Audio Connection Kit specific options')
        self.oscParser = self.generalParser.add_argument_group(
            'OSC', 'Open Sound Control parameters')
        self.setupParsingCLI()

    def performParsingCLI(self):
        """Do what its name sound"""

        self.arguments = self.generalParser.parse_args()
        screenLoggerFormat = logging.Formatter(
            '%(levelname)s:%(filename)s: %(message)s')
        screenLoggerHandler = logging.StreamHandler(stream=sys.stderr)
        screenLoggerHandler.setFormatter(screenLoggerFormat)
        logging.getLogger().addHandler(screenLoggerHandler)
        logging.getLogger().setLevel(getattr(logging, self.arguments.logLevel))
        logging.debug("Command line arguments: " + repr(self.arguments))
        return self.arguments

    def setupParsingCLI(self):
        """Configure the argument parser with some options"""
        # General options
        self.generalParser.add_argument("-l", "--log-level",
                                        choices=['DEBUG', 'INFO',
                                                 'WARNING', 'ERROR', 'CRITICAL'],
                                        dest="logLevel",
                                        default='WARNING',
                                        help="Select logging output level (default WARNING)")
        self.generalParser.add_argument("-v", "--verbose",
                                        action="store_const",
                                        dest="logLevel",
                                        const="DEBUG",
                                        help="Set the log level to DEBUG and be really verbose")
        self.generalParser.add_argument("--length",
                                        type=float,
                                        default=30,
                                        help="Live recording buffer length in seconds (default 30)")

        # OSC
        self.oscParser.add_argument("-p", "--osc-port",
                                    action="store",
                                    type=int,
                                    dest="port",
                                    default=8765,
                                    metavar="OSC_PORT_NUMBER",
                                    help="Define the OSC port to listen to (default 8765)")

        # Portaudio group options
        self.portaudioParser.add_argument("--portaudio",
                                          action="store_const",
                                          dest="audio",
                                          const="portaudio",
                                          help="Define 'portaudio' as the audio server backend and set the supplied number as the main audio interface (see --list-devices to know your audio interface number). If a combination of --portaudio and --jack are used, the last one is the winner")
        self.portaudioParser.add_argument("-a", "--list-portaudio-devices",
                                          action="store_true",
                                          dest="listPortaudioDevices",
                                          help="Print the list of your 'portaudio' system audio interfaces")
        self.portaudioParser.add_argument("-A", "--portaudio-device",
                                          action="store",
                                          type=int,
                                          dest="portaudioDevice",
                                          default=27,
                                          help="Define the 'portaudio' audio interface device by its number (default 27)")

        # Jack group options
        self.jackParser.add_argument("--jack",
                                     action="store_const",
                                     dest="audio",
                                     const="jack",
                                     default="jack",
                                     help="Define 'jack' as the audio server backend (default). If a combination of --portaudio and --jack are used, the last one is the winner")
        self.jackParser.add_argument("--jackname",
                                     action="store",
                                     default="chainsaw",
                                     type=str,
                                     help="Define the name of the 'jack' client. Usefull in conjonction with a 'jack' session management like Qjackctl/PatchBay or Gladish (default 'chainsaw')")
        self.jackParser.add_argument("--sampling-rate",
                                     action="store",
                                     type=int,
                                     dest="samplingRate",
                                     default=48000,
                                     help="Define the audio server sampling rate (default 48000)")



""" Routing configuration for the BCF2000 """
route = {}
for i in range(1, 3):
    route[i] = {
        10: {
            "/chainsaw/instruments/"+str(i)+"/looperTrack": {
                "setMul": ["!"]
            }
        },
        11: {
            "/chainsaw/instruments/"+str(i)+"/looperTrack": {
                "setStart": ["!"]
            }
        },
        12: {
            "/chainsaw/instruments/"+str(i)+"/looperTrack": {
                "setMulController": ["!"]
            }
        },
        13: {
            "/chainsaw/instruments/"+str(i)+"/looperTrack": {
                "setFreqController": ["!"]
            }
        },
        14: {
            "/chainsaw/instruments/"+str(i)+"/looperTrack": {
                "setSharpController": ["!"]
            }
        },
        15: {
            "/chainsaw/instruments/"+str(i): {
                "snapShot": []
            }
        },
        16: {
            "/chainsaw/instruments/"+str(i)+"/looperTrack": {
                "togglePlay": ["!"]
            }
        },
        20: {
            "/chainsaw/Monitor": {
                "setMul": ["!"]
            }
        },
        21: {
            "/chainsaw/instruments/"+str(i)+"/looperTrack": {
                "setDur": ["!"]
            }
        },
        22: {
            "/chainsaw/instruments/"+str(i)+"/looperTrack": {
                "setCenterController": ["!"]
            }
        },
        23: {
            "/chainsaw/instruments/"+str(i)+"/looperTrack": {
                "setShapeController": ["?"]
            }
        },
        24: {
            "/chainsaw/instruments/"+str(i)+"/looperTrack": {
                "setMode": ["?"]
            }
        },
        25: {
            "/chainsaw/instruments/"+str(i): {
                "debug": ["Unnused button, value: ", "?"]
            }
        },
        26: {
            "/chainsaw/instruments/"+str(i)+"/looperTrack": {
                "setMonitor": ["!"]
            }
        },
        # Effect Chain for looperTrack
        30: {
            "/chainsaw/instruments/"+str(i)+"/looperTrack": {
                "setWet": ["!"]
            }
        },
        ## Disto
        31: {
            "/chainsaw/instruments/"+str(i)+"/looperTrack/chain/disto": {
                "setDrive": ["!"]
            }
        },
        41: {
            "/chainsaw/instruments/"+str(i)+"/looperTrack/chain/disto": {
                "setWet": ["!"]
            }
        },
        ## Freeverb
        32: {
            "/chainsaw/instruments/"+str(i)+"/looperTrack/chain/freeverb": {
                "setSize": ["!"]
            }
        },
        42: {
            "/chainsaw/instruments/"+str(i)+"/looperTrack/chain/freeverb": {
                "setWet": ["!"]
            }
        },
        ## FreqShift
        33: {
            "/chainsaw/instruments/"+str(i)+"/looperTrack/chain/freqshift": {
                "setShift": ["!"]
            }
        },
        43: {
            "/chainsaw/instruments/"+str(i)+"/looperTrack/chain/freqshift": {
                "setWet": ["!"]
            }
        },
        ## WaveGuide
        34: {
            "/chainsaw/instruments/"+str(i)+"/looperTrack/chain/waveguide": {
                "setFreq": ["!"]
            }
        },
        44: {
            "/chainsaw/instruments/"+str(i)+"/looperTrack/chain/waveguide": {
                "setWet": ["!"]
            }
        },

        50: {
            "/chainsaw/instruments/"+str(i)+"/particleTrack": {
                "setMul": ["!"]
            }
        },
        51: {
            "/chainsaw/instruments/"+str(i)+"/particleTrack": {
                "setPos": ["!"]
            }
        },
        52: {
            "/chainsaw/instruments/"+str(i)+"/particleTrack": {
                "setMulController": ["!"]
            }
        },
        53: {
            "/chainsaw/instruments/"+str(i)+"/particleTrack": {
                "setFreqController": ["!"]
            }
        },
        54: {
            "/chainsaw/instruments/"+str(i)+"/particleTrack": {
                "setDev": ["!"]
            }
        },
        55: {
            "/chainsaw/instruments/"+str(i): {
                "snapShot": []
            }
        },
        56: {
            "/chainsaw/instruments/"+str(i)+"/particleTrack": {
                "togglePlay": ["!"]
            }
        },

        60: {
            "/chainsaw/instruments/"+str(i)+"/particleTrack": {
                "setDens": ["!"]
            }
        },
        61: {
            "/chainsaw/instruments/"+str(i)+"/particleTrack": {
                "setDur": ["!"]
            }
        },
        62: {
            "/chainsaw/instruments/"+str(i)+"/particleTrack": {
                "setCenterController": ["!"]
            }
        },
        63: {
            "/chainsaw/instruments/"+str(i)+"/particleTrack": {
                "setShapeController": ["?"]
            }
        },
        64: {
            "/chainsaw/instruments/"+str(i)+"/particleTrack": {
                "setEnv": ["?"]
            }
        },
        65: {
            "/chainsaw/instruments/"+str(i): {
                "debug": ["Unnused button, value: ", "?"]
            }
        },
        66: {
            "/chainsaw/instruments/"+str(i)+"/particleTrack": {
                "setMonitor": ["!"]
            }
        },

        # Effect Chain for particleTrack
        70: {
            "/chainsaw/instruments/"+str(i)+"/particleTrack": {
                "setWet": ["!"]
            }
        },
        ## Disto
        71: {
            "/chainsaw/instruments/"+str(i)+"/particleTrack/chain/disto": {
                "setDrive": ["!"]
            }
        },
        81: {
            "/chainsaw/instruments/"+str(i)+"/particleTrack/chain/disto": {
                "setWet": ["!"]
            }
        },
        ## Freeverb
        72: {
            "/chainsaw/instruments/"+str(i)+"/particleTrack/chain/freeverb": {
                "setSize": ["!"]
            }
        },
        82: {
            "/chainsaw/instruments/"+str(i)+"/particleTrack/chain/freeverb": {
                "setWet": ["!"]
            }
        },
        ## FreqShift
        73: {
            "/chainsaw/instruments/"+str(i)+"/particleTrack/chain/freqshift": {
                "setShift": ["!"]
            }
        },
        83: {
            "/chainsaw/instruments/"+str(i)+"/particleTrack/chain/freqshift": {
                "setWet": ["!"]
            }
        },
        ## WaveGuide
        74: {
            "/chainsaw/instruments/"+str(i)+"/particleTrack/chain/waveguide": {
                "setFreq": ["!"]
            }
        },
        84: {
            "/chainsaw/instruments/"+str(i)+"/particleTrack/chain/waveguide": {
                "setWet": ["!"]
            }
        }
    }



# -*- coding: utf-8 -*-
""" This module supply everything's related to instruments."""

import logging
import pyo
from chainsaw.configuration import audioInputs, audioServer, commandLine
from chainsaw.osc import OscNode
from chainsaw.tracks import LooperTrack, ParticleTrack


class Instrument(OscNode, pyo.PyoObject):
    """
    An Instrument is a double track container that hold a LooperTrack and a ParticleTrack
    """

    def __init__(self, name, instrumentInput=audioInputs[1], autoNormalize=True):
        """
        Instrument constructor

        Params:
        name : self
        Instrument name that also define its OSC nodeName.

        instrumentInput : pyo.Input
        The qualified pyo.Input input, or the '0 signaled' global Input as default.

        Doctest:
        >>> i = Instrument('testInstrument')
        >>> i.in_fader is not None
        True
        >>> i.nodeName == 'testInstrument'
        True
        >>> i = Instrument('testInstrument')
        >>> i.nodeName == 'somethingWrong'
        False
        >>> del i
        """
        logging.debug("Creating instrument node named '%s'" % (name))
        OscNode.__init__(self, name)
        pyo.PyoObject.__init__(self)

        self.duration = commandLine.length
        self.in_fader = pyo.InputFader(instrumentInput)
        self.liveBuffer = pyo.NewTable(
            length=self.duration, chnls=1, feedback=0)
        self.recorder = pyo.TableFill(
            input=self.in_fader, table=self.liveBuffer)
        self.fixedBuffer = pyo.NewTable(length=self.liveBuffer.getLength())
        self.setAutoNormalize(autoNormalize)

        self.addChild(LooperTrack("looperTrack", self.fixedBuffer, self))
        self.addChild(ParticleTrack("particleTrack", self.fixedBuffer, self))

        self.out_fader = self.children["looperTrack"].getOut(
        ) + self.children["particleTrack"].getOut()
        self._base_objs = self.out_fader.getBaseObjects()

        self.exposed_to_osc.update(
            {"snapShot": {},
             "togglePlay": {},
             "setPlay": {},
             }
        )

        logging.debug("New instrument named '%s' created with auto normalization '%s'" % (
            name, "on" if self.autoNormalize else "off"))

    def __del__(self):
        """Destructor called when still no reference to object

        See: https://docs.python.org/2/reference/datamodel.html#object.__del__
        """
        logging.info("'%s' Instrument is dying!" % (self.nodeName))

    def view(self):
        """Show something relevant to current instrument, particularly its wavetable"""
        self.fixedBuffer.view(title=commandLine.jackname +
                              " (" + str(commandLine.port) + ")")

    def refreshFixedBuffer(self):
        """
        Update the fixedBuffer buffer with the last live recorded table liveBuffer

        Return:
        True while everything's good
        False otherwise

        Doctest:
        >>> i = Instrument('testInstrument')
        >>> i.in_fader is not None
        True
        >>> i.nodeName == 'testInstrument'
        True
        """
        logging.info("Snapshot the '%s' liveBuffer" % (self.nodeName))
        rotatePosition = int(self.recorder.getCurrentPos())
        logging.debug("Snapshot rotate position for '%s' : %s samples (eq: %s sec)"
                      % (self.nodeName,
                         rotatePosition,
                         rotatePosition / self.getSamplingRate()))
        self.fixedBuffer.copyData(table=self.liveBuffer)
        logging.debug(
            "Dump '%s' liveBuffer to fixedBuffer with 'copyData'" % (self.nodeName))
        self.fixedBuffer.rotate(rotatePosition)
        logging.debug("Rotate '%s' fixedBuffer at %s samples" %
                      (self.nodeName, rotatePosition))
        if self.autoNormalize:
            self.fixedBuffer.normalize()

    def snapShot(self):
        """Wrapper to refreshFixedBuffer"""
        self.refreshFixedBuffer()

    def setAutoNormalize(self, value=True):
        """Define the auto normalization flag

        Params:
        value : bool
        Th flag vaue. Default to True.
        """
        self.autoNormalize = True if value is None else value

    def setInput(self, newInput, fadetime=0.05):
        """
        Set the Instrument input to another one

        Params:
        newInput : pyo.PyoObject or int
        New input to use for live recording.
        If int, corresponding jack audio input wil be used

        fadetime : float
        Crossfade time between old and new input. Defaults to 0.05.
        """
        logging.debug("Set '%s' audio input signal to '%s'" %
                      (self.nodeName, newInput))
        if isinstance(newInput, int) and newInput in range(len(audioInputs)):
            self.in_fader.setInput(audioInputs[newInput], float(fadetime))
        elif isinstance(newInput, pyo.PyoObject) or hasattr(newInput, "stream"):
            self.in_fader.setInput(newInput, float(fadetime))

    def play(self, dur=0, delay=0):
        """Start processing without sending samples to output. This method is called automatically at the object creation.

        This method returns self, allowing it to be applied at the object creation.

        Params:
        dur : float
        Duration, in seconds, of the object’s activation. The default is 0 and means infinite duration.

        delay : float
        Delay, in seconds, before the object’s activation. Defaults to 0.
        """
        self.out_fader.play(dur, delay)
        return pyo.PyoObject.play(self, dur, delay)

    def stop(self):
        """Stop processing.

        This method returns self, allowing it to be applied at the object creation.
        """
        self.out_fader.stop()
        return pyo.PyoObject.stop(self)

    def out(self, chnl=0, inc=1, dur=0, delay=0):
        """Start processing and send samples to audio output beginning at chnl.

        This method returns self, allowing it to be applied at the object creation.

        Params:
        chnl : int
        Physical output assigned to the first audio stream of the object. Defaults to 0.

        inc : int
        Output channel increment value. Defaults to 1.

        dur : float
        Duration, in seconds, of the object’s activation. The default is 0 and means infinite duration.

        delay : float
        Delay, in seconds, before the object’s activation. Defaults to 0.
        """
        logging.info("Instrument '%s' sends its signal to output #%s" %
                     (self.nodeName, chnl))
        self.out_fader.play(dur, delay)
        return pyo.PyoObject.out(self, chnl, inc, dur, delay)

    def togglePlay(self):
        """
        Toggle Intrument's state between [playing] and [stopped] based on current state
        """
        if self.isPlaying():
            self.stop()
        else:
            self.play()

    def setPlay(self, play=False):
        """
        Set Intrument's state between [playing] and [stopped]
        """
        if play:
            self.play()
        else:
            self.stop()


if __name__ == "__main__":
    import doctest
    fails, total = doctest.testmod(verbose=True)
    if total > 0 and fails == 0:
        instrument = Instrument("1")
        patLive = pyo.Pattern(instrument.liveBuffer.refreshView, 0.05).play()
        patFixed = pyo.Pattern(instrument.fixedBuffer.refreshView, 0.05).play()
        instrument.liveBuffer.view()
        instrument.fixedBuffer.view()
        audioServer.start()
        audioServer.gui(locals())



#!/usr/bin/python3
# -*- coding: utf-8 -*-
""" Graph the routing using Graphviz """

import logging
import pygraphviz as pgv
import cli

def main():
    """ Main logic"""
    command_line = cli.CLI().performParsingCLI()
    logging.debug("")
    route = command_line.routeModule.route
    graph = pgv.AGraph(name="MIDI2OSCrouting",
                       label="Chainsaw routing MIDI to OSC",
                       strict=True,
                       directed=True)
    graph.node_attr['shape'] = 'record'

    def recursive_elements(_arbre, _osc_string, _channel, _control_change, _function=None):
        """ recursively create a dictionary from the route mapping """
        if len(_osc_string.split('/')) == 1:
            if isinstance(_arbre, dict):
                if _osc_string not in _arbre.keys():
                    _arbre[_osc_string] = dict()
            if _function not in _arbre[_osc_string]:
                _arbre[_osc_string][_function] = {'linker': '%s%s%s' % (_function,
                                                                        _channel,
                                                                        _control_change),
                                                  'chan': _channel,
                                                  'cc': _control_change}
        else:
            element = _osc_string.split('/', maxsplit=1)[0]
            if element not in _arbre.keys():
                _arbre[element] = dict()
            recursive_elements(_arbre[element],
                               _osc_string.split('/', maxsplit=1)[1],
                               _channel,
                               _control_change,
                               _function)
        return

    osc_structure = dict()
    for channel in route:
        for control_change in route[channel]:
            for target in route[channel][control_change]:
                for function in route[channel][control_change][target]:
                    recursive_elements(osc_structure, target[1:], channel, control_change, function)

    for channel in route:
        node_name = 'chnl_' + str(channel)
        channel_name = 'Chan: ' + str(channel)
        node_label = channel_name + '|{'
        cc_counter = 0
        for control_change in route[channel]:
            cc_source_name = 'cc%s' % (control_change)
            cc_name = '<%s> cc:%s' % (cc_source_name, control_change)
            if cc_counter == 0:
                node_label = '%s %s' % (node_label, cc_name)
            else:
                node_label = '%s | %s' % (node_label, cc_name)
            for target in route[channel][control_change]:
                for function in route[channel][control_change][target]:
                    graph.add_node(function)
                    graph.add_edge(node_name, function, tailport=cc_source_name, label='t')
            cc_counter += 1
        node_label = node_label + '}'
        graph.add_node(node_name, label=node_label)

    graph.layout(prog='fdp')
    graph.draw('route.pdf')

if __name__ == "__main__":
    main()
